/*
 * Procesor de efecte pentru chitara
 * Fera Daniel <feradaniel21@gmail.com>
 * Grupa 333CC
 */
#define SGN_MIN -512
#define SGN_MAX 512
#define IN_WAKEUP 10
#define TIMER_1HZ 15625
#define MAX_OUTPUT_HISTORY 800

#include "lcd.h"
#include "menu.h"
#include "twi.h"

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/atomic.h>
#include <math.h>

uint8_t ui_check = 0;

/* Effects presets & settings */

enum effects {
    EFFECT_DISTORTION = 0,
    EFFECT_TREMOLO,
    EFFECT_FLANGER,
    EFFECT_ECHO,
    EFFECT_COMPRESSION
};

uint16_t active_effects;

//DISTORTION
float es_distortion_depth = 0.5f;
float es_distortion_gain = 20.0f;

//TREMOLO
float es_tremolo_depth = 0.8f;
uint16_t es_tremolo_rate_Hz = 2;

//FLANGER
float es_flanger_delay = 0.02f;
float es_flanger_depth = 0.4f;
uint16_t es_flanger_rate_Hz = 1;

//ECHO
// number between 0 and MAX_OUTPUT_HISTORY - 1
uint16_t es_echo_delay = 400;
float es_echo_depth = 0.7f;

//COMPRESSION
float es_compression_knee_treshold = 0.1f;
float es_compression_kmin = 0.10f;
float es_compression_alpha_release = 0.7f;
float es_compression_alpha_attack = 0.7f;

/* ------------------------------------ */

int16_t output_dac;
int16_t read_adc;

int16_t output_history[MAX_OUTPUT_HISTORY];
int16_t output_history_last = 0;
int16_t output_history_first = 1;

#define OUTPUT_HISTORY_ADD(x) do { \
    output_history[output_history_last] = x; \
    output_history_last = (output_history_last + 1) % MAX_OUTPUT_HISTORY; \
    output_history_first = (output_history_first + 1) % MAX_OUTPUT_HISTORY; \
    } while(0);

uint16_t timer0_counter = 0;
uint16_t timer0_top = 0;
#define TCNT0E ((timer0_counter + TCNT0) % (timer0_top + 1))

ISR(TIMER0_COMPA_vect)
{
    timer0_counter = (timer0_counter + 0xFF) % (timer0_top + 1);
}

// Pin change interrupt handler
ISR(PCINT1_vect)
{
    ui_check = 1;
}

// keep signal in range [-SGN_MIN, SGN_MAX]
float adjust_amplitude(float x)
{
    if (x < -1.0f)
        return -1.0f;
    if (x > 1.0f)
        return 1.0f;
    return x;
}

float apply_distortion(float x)
{
    return (1-es_distortion_depth) * x + es_distortion_depth * tanh(es_distortion_gain * x);
}

float apply_tremolo(float x)
{
    float modulation = sin((2 * M_PI / OCR1A) * TCNT1);
    modulation = (1 - es_tremolo_depth) + es_tremolo_depth * modulation * modulation;

    return x * modulation;
}

float apply_flanger(float x)
{ 
    float max_sample_delay = es_flanger_delay * MAX_OUTPUT_HISTORY;
    float current_sin = fabs(sin((2 * M_PI / timer0_top) * TCNT0E));
    int16_t current_delay = (int16_t)ceil(current_sin * max_sample_delay);

    float out_delayed = (float)output_history[(output_history_last + (MAX_OUTPUT_HISTORY - 1) - current_delay) % MAX_OUTPUT_HISTORY];
    out_delayed /= SGN_MAX;

    return (1 - es_flanger_depth) * x +
        es_flanger_depth * out_delayed;
}

float apply_echo(float x)
{
    float out_delayed = (float)output_history[(output_history_last + (MAX_OUTPUT_HISTORY - 1) - es_echo_delay) % MAX_OUTPUT_HISTORY];
    out_delayed /= SGN_MAX;

    return (1 - es_echo_depth) * x + es_echo_depth * out_delayed;
}

float apply_compression(float x)
{
    const float slope = 1.0f / 5.0f;
    static float compressor_gain = 1.0f;
    static float state = 0.0f;

    const float max_abs = fmax(x, es_compression_kmin);
    const float max_abs_dB = log(max_abs);
    const float overshoot = max_abs_dB - es_compression_knee_treshold;
    const float rect = fmax(overshoot, 0.0f);
    const float cv = rect * slope;
    const float prev_state = state;
    
    if (cv <= state)
        state = es_compression_alpha_attack * state + (1 - es_compression_alpha_attack) * cv;
    else
        state = es_compression_alpha_release * state + (1 - es_compression_alpha_release) * cv;

    compressor_gain *= exp(state - prev_state);

    return x * compressor_gain;
}

float process_signal(float x)
{
    // TODO find right chain order

    if (active_effects & _BV(EFFECT_COMPRESSION))
        x = apply_compression(x);

    if (active_effects & _BV(EFFECT_DISTORTION))
        x = apply_distortion(x);

    if (active_effects & _BV(EFFECT_TREMOLO))
        x = apply_tremolo(x);

    if (active_effects & _BV(EFFECT_FLANGER))
        x = apply_flanger(x);
    
    if (active_effects & _BV(EFFECT_ECHO))
        x = apply_echo(x);
    
    return adjust_amplitude(x);
}

// ADC conversion complete handler
ISR(ADC_vect)
{
    // gather from ADC
    read_adc = ADC;

    // Twos Complement converter
    if (read_adc & _BV(9)) {
        //negative number
        read_adc = read_adc | ~(_BV(10) - 1);
    }

    // process normalized signal [-1, 1]
    output_dac = (int16_t)(process_signal(((float)read_adc) / SGN_MAX) * SGN_MAX);
    OUTPUT_HISTORY_ADD(output_dac);
    
    output_dac = output_dac + SGN_MAX;
    DAC_output(output_dac);

    if (ui_check) {
        // check for menu key inputs
        if ((PINB & _BV(PB0)) == 0) {
            // Left (Back)
            MENU_updatestate(MENU_INPUT_BACK);
        } else if ((PINB & _BV(PB2)) == 0) {
            // Down
            MENU_updatestate(MENU_INPUT_DOWN);
        } else if ((PINB & _BV(PB1)) == 0) {
            // Up
            MENU_updatestate(MENU_INPUT_UP);
        } else if ((PINB & _BV(PB3)) == 0) {
            MENU_updatestate(MENU_INPUT_OK);
        }
        ui_check = 0;
    }
}

void loop()
{
    for(;;);
}

void SignalInput_init()
{
     // Initialize ADC
    ADMUX = 0
        // Internal 2.56V Voltage Reference with external capacitor at AREF pin
        | _BV(REFS1)
        | _BV(REFS0)

        // Measure ADC1 - ADC0 with 10x gain
        | 0b01001;

    ADCSRA = 0
        // Activate ADC interrupt
        | _BV(ADIE)

        // Activate ADC Auto Triggering
        | _BV(ADATE)

        // TODO: try more prescaler settings
        // Set ADC Prescaler to 64 => f(ADC) = 16MHz / 64 = 250 kHz
        | 0b110;

    // ADC Free Running Mode - Auto Trigger
    ADCSRB = 0;

    // Disable digital input buffer for ADC pins (ADC0, ADC1)
    DIDR0 = 0b00000011;

    // Enable ADC
    ADCSRA |= _BV(ADEN);
    
    // First ADC conversion
    ADCSRA |= _BV(ADSC);
}

void SignalOutput_init()
{
    // initialzie external DAC
    TWI_init();

    // initialize timer1 to be used with the tremolo effect
    TCCR1A = 0;
    TCCR1B = 0
        | _BV(WGM12)
        | _BV(CS12)
        | _BV(CS10);
    
    // TODO settings
    OCR1A = TIMER_1HZ / 2;

    //initialize timer0 to be used with the flanger effect
    // TODO settings
    timer0_top = TIMER_1HZ;
    TCCR0A = 0
        | _BV(WGM01);
    TCCR0B = 0
        | _BV(CS02)
        | _BV(CS00);
    TIMSK0 = 0
        | _BV(OCIE0A);
    OCR0A = 0xFF;
}

void UI_init()
{
     // Activate pull-up resistors for control buttons
    PORTB |= _BV(PB0) | _BV(PB1) | _BV(PB2) | _BV(PB3);

    // Activate interrupts for PB0..3
    PCICR  = (1 << PCIE1);
    PCMSK1 = (1 << PCINT8) | (1 << PCINT9)
        | (1 << PCINT10) | (1 << PCINT11);
}

int main()
{
    // Activate global interrupts
    sei();

    /* Initialize */
    MENU_init();
    UI_init();
    SignalOutput_init();
    SignalInput_init();

    /* Main Loop */
    loop();

    return 0;
}
